@@ -6,21 +6,21 @@
     This module is treated as private API.
     Users should not need to use this module directly.
 """
-import re
+
+from __future__ import annotations
+
 import functools
 import operator
+import re
 import typing
 import warnings
-from packaging.version import Version
 
 import marshmallow
 from marshmallow.orderedset import OrderedSet
-
-
-RegexType = type(re.compile(""))
+from packaging.version import Version
 
 # marshmallow field => (JSON Schema type, format)
-DEFAULT_FIELD_MAPPING = {
+DEFAULT_FIELD_MAPPING: dict[type, tuple[str | None, str | None]] = {
     marshmallow.fields.Integer: ("integer", None),
     marshmallow.fields.Number: ("number", None),
     marshmallow.fields.Float: ("number", None),
@@ -38,6 +38,9 @@ DEFAULT_FIELD_MAPPING = {
     marshmallow.fields.Field: (None, None),
     marshmallow.fields.Raw: (None, None),
     marshmallow.fields.List: ("array", None),
+    marshmallow.fields.IP: ("string", "ip"),
+    marshmallow.fields.IPv4: ("string", "ipv4"),
+    marshmallow.fields.IPv6: ("string", "ipv6"),
 }
 
 
@@ -88,7 +91,7 @@ _VALID_PREFIX = "x-"
 class FieldConverterMixin:
     """Adds methods for converting marshmallow fields to an OpenAPI properties."""
 
-    field_mapping = DEFAULT_FIELD_MAPPING
+    field_mapping: dict[type, tuple[str | None, str | None]] = DEFAULT_FIELD_MAPPING
     openapi_version: Version
 
     def init_attribute_functions(self):
@@ -100,7 +103,6 @@ class FieldConverterMixin:
             self.field2choices,
             self.field2read_only,
             self.field2write_only,
-            self.field2nullable,
             self.field2range,
             self.field2length,
             self.field2pattern,
@@ -111,6 +113,8 @@ class FieldConverterMixin:
             self.list2properties,
             self.dict2properties,
             self.timedelta2properties,
+            self.datetime2properties,
+            self.field2nullable,
         ]
 
     def map_to_openapi_type(self, field_cls, *args):
@@ -149,7 +153,7 @@ class FieldConverterMixin:
             merged with the return values of all other attribute functions called on the field.
             User added attribute functions will be called after all built-in attribute
             functions in the order they were added. The merged results of all
-            previously called attribute functions are accessable via the `ret`
+            previously called attribute functions are accessible via the `ret`
             argument.
         """
         bound_func = func.__get__(self)
@@ -191,10 +195,9 @@ class FieldConverterMixin:
                 break
         else:
             warnings.warn(
-                "Field of type {} does not inherit from marshmallow.Field.".format(
-                    type(field)
-                ),
+                f"Field of type {type(field)} does not inherit from marshmallow.Field.",
                 UserWarning,
+                stacklevel=2,
             )
             type_, fmt = "string", None
 
@@ -239,7 +242,7 @@ class FieldConverterMixin:
         attributes = {}
 
         comparable = [
-            validator.comparable  # type:ignore
+            validator.comparable
             for validator in field.validators
             if hasattr(validator, "comparable")
         ]
@@ -247,13 +250,18 @@ class FieldConverterMixin:
             attributes["enum"] = comparable
         else:
             choices = [
-                OrderedSet(validator.choices)  # type:ignore
+                OrderedSet(validator.choices)
                 for validator in field.validators
                 if hasattr(validator, "choices")
             ]
             if choices:
                 attributes["enum"] = list(functools.reduce(operator.and_, choices))
 
+        if field.allow_none:
+            enum = attributes.get("enum")
+            if enum is not None and None not in enum:
+                attributes["enum"].append(None)
+
         return attributes
 
     def field2read_only(
@@ -293,9 +301,25 @@ class FieldConverterMixin:
             if self.openapi_version.major < 3:
                 attributes["x-nullable"] = True
             elif self.openapi_version.minor < 1:
-                attributes["nullable"] = True
+                if "$ref" in ret:
+                    attributes["anyOf"] = [
+                        {"type": "object", "nullable": True},
+                        {"$ref": ret.pop("$ref")},
+                    ]
+                elif "allOf" in ret:
+                    attributes["anyOf"] = [
+                        *ret.pop("allOf"),
+                        {"type": "object", "nullable": True},
+                    ]
+                else:
+                    attributes["nullable"] = True
             else:
-                attributes["type"] = [*make_type_list(ret.get("type")), "null"]
+                if "$ref" in ret:
+                    attributes["anyOf"] = [{"$ref": ret.pop("$ref")}, {"type": "null"}]
+                elif "allOf" in ret:
+                    attributes["anyOf"] = [*ret.pop("allOf"), {"type": "null"}]
+                else:
+                    attributes["type"] = [*make_type_list(ret.get("type")), "null"]
         return attributes
 
     def field2range(self, field: marshmallow.fields.Field, ret) -> dict:
@@ -320,7 +344,12 @@ class FieldConverterMixin:
             if set(make_type_list(ret.get("type"))) & {"number", "integer"}
             else ("x-minimum", "x-maximum")
         )
-        return make_min_max_attributes(validators, min_attr, max_attr)
+
+        # Serialize min/max values with the field to which the validator is applied
+        return {
+            k: field._serialize(v, None, None)
+            for k, v in make_min_max_attributes(validators, min_attr, max_attr).items()
+        }
 
     def field2length(
         self, field: marshmallow.fields.Field, **kwargs: typing.Any
@@ -348,9 +377,7 @@ class FieldConverterMixin:
         max_attr = "maxItems" if is_array else "maxLength"
 
         equal_list = [
-            validator.equal  # type:ignore
-            for validator in validators
-            if validator.equal is not None  # type:ignore
+            validator.equal for validator in validators if validator.equal is not None
         ]
         if equal_list:
             return {min_attr: equal_list[0], max_attr: equal_list[0]}
@@ -372,16 +399,17 @@ class FieldConverterMixin:
         regex_validators = (
             v
             for v in field.validators
-            if isinstance(getattr(v, "regex", None), RegexType)
+            if isinstance(getattr(v, "regex", None), re.Pattern)
         )
         v = next(regex_validators, None)
         attributes = {} if v is None else {"pattern": v.regex.pattern}  # type:ignore
 
         if next(regex_validators, None) is not None:
             warnings.warn(
-                "More than one regex validator defined on {} field. Only the "
-                "first one will be used in the output spec.".format(type(field)),
+                f"More than one regex validator defined on {type(field)} field. Only the "
+                "first one will be used in the output spec.",
                 UserWarning,
+                stacklevel=2,
             )
 
         return attributes
@@ -437,7 +465,17 @@ class FieldConverterMixin:
             field, marshmallow.fields.Pluck
         ):
             schema_dict = self.resolve_nested_schema(field.schema)  # type:ignore
-            if ret and "$ref" in schema_dict:
+            if (
+                ret
+                and "$ref" in schema_dict
+                and (
+                    self.openapi_version.major < 3
+                    or (
+                        self.openapi_version.major == 3
+                        and self.openapi_version.minor == 0
+                    )
+                )
+            ):
                 ret.update({"allOf": [schema_dict]})
             else:
                 ret.update(schema_dict)
@@ -485,6 +523,8 @@ class FieldConverterMixin:
             value_field = field.value_field
             if value_field:
                 ret["additionalProperties"] = self.field2property(value_field)
+            else:
+                ret["additionalProperties"] = {}
         return ret
 
     def timedelta2properties(self, field, **kwargs: typing.Any) -> dict:
@@ -514,6 +554,57 @@ class FieldConverterMixin:
             else:
                 choices = (m.value for m in field.enum)
             ret["enum"] = [field.field._serialize(v, None, None) for v in choices]
+            if field.allow_none and None not in ret["enum"]:
+                ret["enum"].append(None)
+        return ret
+
+    def datetime2properties(self, field, **kwargs: typing.Any) -> dict:
+        """Return a dictionary of properties from :class:`DateTime <marshmallow.fields.DateTime` fields.
+
+        :param Field field: A marshmallow field.
+        :rtype: dict
+        """
+        ret = {}
+        if isinstance(field, marshmallow.fields.DateTime) and not isinstance(
+            field, marshmallow.fields.Date
+        ):
+            if field.format == "iso" or field.format is None:
+                # Will return { "type": "string", "format": "date-time" }
+                # as specified inside DEFAULT_FIELD_MAPPING
+                pass
+            elif field.format == "rfc":
+                ret = {
+                    "type": "string",
+                    "format": None,
+                    "example": "Wed, 02 Oct 2002 13:00:00 GMT",
+                    "pattern": r"((Mon|Tue|Wed|Thu|Fri|Sat|Sun), ){0,1}\d{2} "
+                    + r"(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) \d{4} \d{2}:\d{2}:\d{2} "
+                    + r"(UT|GMT|EST|EDT|CST|CDT|MST|MDT|PST|PDT|(Z|A|M|N)|(\+|-)\d{4})",
+                }
+            elif field.format == "timestamp":
+                ret = {
+                    "type": "number",
+                    "format": "float",
+                    "example": "1676451245.596",
+                    "min": "0",
+                }
+            elif field.format == "timestamp_ms":
+                ret = {
+                    "type": "number",
+                    "format": "float",
+                    "example": "1676451277514.654",
+                    "min": "0",
+                }
+            else:
+                ret = {
+                    "type": "string",
+                    "format": None,
+                    "pattern": (
+                        field.metadata["pattern"]
+                        if field.metadata.get("pattern")
+                        else None
+                    ),
+                }
         return ret
 
 
